@@ -3,13 +3,22 @@ import sys
 from contextlib import AbstractContextManager
 from pathlib import PurePath
 from platform import platform
+from typing import Any, BinaryIO, Callable, Dict, IO, Iterable, List, Optional, Tuple, Union
 
 import click
 import polling2
 import requests
+import urllib3
 from requests import Response
-from typing import Any, BinaryIO, Callable, Dict, IO, Iterable, List, Optional, Tuple, Union
+from tenacity import (
+    retry,
+    stop_after_attempt,
+    stop_after_delay,
+    wait_fixed,
+    retry_if_exception_type,
+)
 
+from .sideloading import to_sideloads, Sideload
 from rossum import __version__, CTX_PROFILE, CTX_DEFAULT_PROFILE
 from rossum.configure import get_credential
 from . import (
@@ -27,7 +36,13 @@ from . import (
 
 
 class RossumException(click.ClickException):
-    pass
+    def __init__(self, message: str, response: Response = None):
+        super().__init__(message)
+        self.response = response
+
+    @property
+    def status_code(self):
+        return self.response.status_code if self.response else 400
 
 
 RequestsFiles = Dict[str, Tuple[Optional[str], Union[IO[bytes], BinaryIO, str]]]
@@ -45,7 +60,25 @@ class APIClient(AbstractContextManager):
         use_api_version: bool = True,
         auth_using_token: bool = True,
         max_token_lifetime: Optional[int] = None,
+        retry_logic_rules: Optional[Dict] = None,
     ):
+        """The APIClient for communication with Rossum API.
+
+        :param context Used for switching profile. Send None for usual requests done for your default profile.
+        :param url Rossum API URL. Leave None for communication with the currently used publicly available Rossum API.
+        :param user Your username.
+        :param password Your password.
+        :param use_api_version Leave True if you want to use the latest API version. If set to false,
+        specify the Rossum API URL including its version in the url parameter explicitly.
+        :param auth_using_token To avoid using login request for each call, leave set to True.
+        :param max_token_lifetime Set custom max token lifetime in seconds. Default is the maximum lifetime: 583200s
+        or until the Rossum client logs out. Logging out is made when CLI is exited or if RossumClient() is used
+        with a with statement.
+        :param retry_logic_rules Pass logic rules for built-in retry mechanism that is called when it is currently
+        not possible to communicate with Rossum API. The default shape of the dictionary is:
+        {"attempts": 3, "wait_s": 5}. "attempts" key states the number of retry attempts. "wait_s" is time in seconds
+        the APIClient will wait before retrying again.
+        """
         self._url = url
         self._user = user
         self._password = password
@@ -57,6 +90,8 @@ class APIClient(AbstractContextManager):
         self.token: Optional[str] = None
         self.timeout: Optional[float] = None
 
+        self._retry_logic_rules = self.get_retry_logic(retry_logic_rules)
+
     def __exit__(self, exc_type, exc_value, traceback) -> None:
         self.logout()
 
@@ -85,16 +120,42 @@ class APIClient(AbstractContextManager):
             self._url = f'{_url}{"/v1" if self._use_api_version else ""}'
         return self._url
 
+    @staticmethod
+    def get_retry_logic(retry_logic_rules: Optional[Dict]):
+        retry_logic_rules = retry_logic_rules or {}
+        attempts_no = retry_logic_rules.get("attempts", 3)
+        wait_s = retry_logic_rules.get("wait_s", 5)
+        retry_logic = {
+            "reraise": True,
+            "stop": (stop_after_attempt(attempts_no) | stop_after_delay(55)),
+            "wait": wait_fixed(wait_s),
+            "retry": (
+                retry_if_exception_type(requests.exceptions.ProxyError)
+                | retry_if_exception_type(requests.exceptions.ConnectionError)
+                | retry_if_exception_type(urllib3.exceptions.NewConnectionError)
+                | retry_if_exception_type(urllib3.exceptions.ConnectTimeoutError)
+            ),
+        }
+        return retry_logic
+
+    def _login_to_api(self, login_data: Dict) -> Response:
+        return requests.post(
+            f"{self.url}/auth/login", json=login_data, timeout=self.timeout, headers=HEADERS
+        )
+
     def get_token(self) -> str:
         # self.post cannot be used as it is dependent on self.get_token().
         login_data: Dict[str, Union[str, int]] = {"username": self.user, "password": self.password}
         if self._max_token_lifetime:
             login_data["max_token_lifetime_s"] = self._max_token_lifetime
-        response = requests.post(f"{self.url}/auth/login", json=login_data, headers=HEADERS)
+        retry_request = retry(**self._retry_logic_rules)(self._login_to_api)
+        response = retry_request(login_data)
         if response.status_code == 401:
-            raise RossumException(f"Login failed with the provided credentials.")
+            raise RossumException("Login failed with the provided credentials.", response=response)
         elif not response.ok:
-            raise RossumException(f"Invalid response [{response.url}]: {response.text}")
+            raise RossumException(
+                f"Invalid response [{response.url}]: {response.text}", response=response
+            )
 
         return response.json()["key"]
 
@@ -142,9 +203,12 @@ class APIClient(AbstractContextManager):
     def _request_url(
         self, method: str, url: str, query: dict = None, expected_status_code: int = 200, **kwargs
     ) -> Response:
-        response = self._do_request(method, url, query, **kwargs)
+        retry_request = retry(**self._retry_logic_rules)(self._do_request)
+        response = retry_request(method, url, query, **kwargs)
         if response.status_code != expected_status_code:
-            raise RossumException(f"Invalid response [{response.url}]: {response.text}")
+            raise RossumException(
+                f"Invalid response [{response.url}]: {response.text}", response=response
+            )
         return response
 
     def delete(self, to_delete: Dict[str, str], verbose: int = 0, item: str = "annotation") -> None:
@@ -166,30 +230,63 @@ class APIClient(AbstractContextManager):
         query: Optional[Dict[str, Any]] = None,
         *,
         key: str = "results",
+        sideloads: Optional[Iterable[Union[Sideload, str]]] = None,
     ) -> Tuple[List[Dict[str, Any]], int]:
+        if query is None:
+            query = {}
+
+        sideloads = to_sideloads(sideloads)
+
+        if sideloads:
+            if "sideload" in query:
+                raise RossumException("sideloading cannot be specified both in query and sideloads")
+            query = query.copy()
+            for sideload in sideloads:
+                sideload.setup_query(query)
+
         response = self.get(path, query)
         response_dict = response.json()
 
-        res = response_dict[key]
+        res = response_dict
         next_page = response_dict["pagination"]["next"]
 
         while next_page:
             response = self.get_url(next_page)
             response_dict = response.json()
 
-            res.extend(response_dict[key])
+            for k, v in response_dict.items():
+                if k != "pagination":
+                    res.setdefault(k, []).extend(v)
             next_page = response_dict["pagination"]["next"]
 
-        return res, response_dict["pagination"]["total"]
+        if sideloads:
+            self._resolve_sideloads(res, sideloads)
+
+        return res[key], response_dict["pagination"]["total"]
 
     def _sideload(
         self, objects: List[dict], sideloads: Optional[Iterable[APIObject]] = None
     ) -> List[dict]:
+        response = {"results": objects}
         for sideload in sideloads or []:
             sideloaded, _ = self.get_paginated(sideload)
-            sideloaded_dicts = {
-                sideloaded_dict["url"]: sideloaded_dict for sideloaded_dict in sideloaded
-            }
+            response[sideload.plural] = sideloaded
+        if sideloads:
+            self._resolve_sideloads(response, to_sideloads(sideloads))
+        return response["results"]
+
+    def _resolve_sideloads(self, response: dict, sideloads: Iterable[Sideload]) -> None:
+        """
+        Resolve the dependency injections of sideloaded objects to corresponding
+        results objects in the `response`.
+
+        :param response Response object as returned from Rossum app for any
+        paginated GET-like request.
+        :param sideloads Optional list of object types to resolve dependency
+        injection for.
+        """
+        for sideload in sideloads or []:
+            sideloaded_dicts = sideload.get_mapping(response.get(sideload.plural, []))
 
             def inject_sideloaded(obj: dict) -> dict:
                 try:
@@ -204,8 +301,8 @@ class APIClient(AbstractContextManager):
                     obj[sideload.singular] = sideloaded_dicts.get(url, {})
                 return obj
 
-            objects = [inject_sideloaded(o) for o in objects]
-        return objects
+            for obj in response["results"]:
+                inject_sideloaded(obj)
 
     @property
     def _authentication(self) -> dict:
@@ -264,11 +361,14 @@ class RossumClient(APIClient):
         self,
         sideloads: Optional[Iterable[APIObject]] = None,
         *,
+        any_of_ids: Optional[Iterable[int]] = None,
         workspace: Optional[int] = None,
         users: Optional[Iterable[int]] = None,
         hooks: Optional[Iterable[int]] = None,
     ) -> List[dict]:
         query: Dict[str, Any] = {}
+        if any_of_ids:
+            query["id"] = ",".join(str(qid) for qid in any_of_ids)
         if workspace:
             query[WORKSPACES.singular] = workspace
         if users:
@@ -339,6 +439,21 @@ class RossumClient(APIClient):
             raise RossumException("Annotation ID wasn't specified.")
         return get_json(self.get(f"{ANNOTATIONS}/{id_}"))
 
+    def get_annotations(
+        self,
+        *,
+        queue: Optional[int] = None,
+        status: Optional[Iterable[str]] = None,
+        sideloads: Optional[Iterable[Union[Sideload, str]]] = None,
+    ):
+        query = {}
+        if queue is not None:
+            query["queue"] = str(queue)
+        if status:
+            query["status"] = ",".join(status)
+        annotations, _ = self.get_paginated(ANNOTATIONS, query=query, sideloads=sideloads)
+        return annotations
+
     def poll_annotation(
         self, annotation: int, check_success: Callable, max_retries=120, sleep_secs=5
     ) -> dict:
@@ -400,25 +515,27 @@ class RossumClient(APIClient):
         return get_json(self.post("queues", data))
 
     def create_inbox(
-        self, name: str, email_prefix: Optional[str], bounce_email: Optional[str], queue_url: str
+        self,
+        name: str,
+        email_prefix: Optional[str],
+        bounce_email: Optional[str],
+        queue_url: str,
+        email: Optional[str] = None,
     ) -> dict:
-        if not (email_prefix and bounce_email):
+        if not (email_prefix or email):
             raise RossumException(
-                "Inbox cannot be created without both bounce email and email prefix specified."
+                "Inbox cannot be created without email prefix or email specified."
             )
-
-        return get_json(
-            self.post(
-                "inboxes",
-                data={
-                    "name": name,
-                    "email_prefix": email_prefix,
-                    "bounce_email_to": bounce_email,
-                    "bounce_unprocessable_attachments": True,
-                    "queues": [queue_url],
-                },
-            )
-        )
+        inbox_creation_data = {
+            "name": name,
+            "email_prefix": email_prefix,
+            "bounce_email_to": bounce_email,
+            "bounce_unprocessable_attachments": True if bounce_email else False,
+            "queues": [queue_url],
+        }
+        if email:
+            inbox_creation_data["email"] = email
+        return get_json(self.post("inboxes", data=inbox_creation_data))
 
     def create_user(
         self,
@@ -428,13 +545,14 @@ class RossumClient(APIClient):
         password: str,
         group: str,
         locale: str,
+        email: str = "",
     ) -> dict:
         return get_json(
             self.post(
                 USERS,
                 data={
                     "username": username,
-                    "email": username,
+                    "email": email if email else username,
                     "organization": organization,
                     "password": password,
                     "groups": [g["url"] for g in self.get_groups(group_name=group)],
@@ -484,6 +602,11 @@ class RossumClient(APIClient):
         events: List[str],
         sideload: List[str],
         config: Dict,
+        run_after: List[str] = None,
+        metadata: Optional[Dict] = None,
+        token_owner: Optional[str] = "",
+        test: Optional[Dict] = None,
+        **kwargs: Any,
     ) -> dict:
 
         data = {
@@ -494,16 +617,40 @@ class RossumClient(APIClient):
             "events": events,
             "sideload": sideload,
             "config": config,
+            "metadata": metadata or {},
+            "run_after": run_after or [],
+            "test": test or {},
         }
+        if token_owner:
+            data["token_owner"] = token_owner
+        if kwargs:
+            data = {**data, **kwargs}
         return get_json(self.post("hooks", data))
 
     def upload_document(
-        self, id_: int, file: str, filename_overwrite: str = "", values: Dict[str, str] = None
+        self,
+        id_: int,
+        file: Optional[str] = "",
+        filename_overwrite: str = "",
+        values: Dict[str, str] = None,
+        metadata: Optional[Dict] = None,
+        file_bytes: Optional[bytes] = None,
     ) -> dict:
-        filename = PurePath(filename_overwrite).name or PurePath(file).name
-        files: RequestsFiles = {"content": (filename, open(f"{file}", "rb"))}
+        if not (file or file_bytes):
+            raise RossumException("No file(s) to be sent.")
+        if file_bytes and not filename_overwrite:
+            raise RossumException(
+                "Need to pass filename_overwrite parameter to state the file name for the uploaded document."
+            )
+        if file:
+            filename = PurePath(filename_overwrite).name or PurePath(file).name
+            files: RequestsFiles = {"content": (filename, open(f"{file}", "rb"))}
+        else:
+            files: RequestsFiles = {"content": (filename_overwrite, file_bytes)}  # type: ignore
         if values is not None:
             files["values"] = (None, json.dumps(values))
+        if metadata:
+            files["metadata"] = (None, json.dumps(metadata))
         return get_json(self.post(f"queues/{id_}/upload", files=files))
 
     def set_metadata(self, object_type: APIObject, object_id: int, metadata: Dict[str, Any]):
@@ -523,14 +670,18 @@ def get_json(response: Response) -> dict:
     try:
         return response.json()
     except ValueError as e:
-        raise RossumException(f"Invalid JSON [{response.url}]: {response.text}") from e
+        raise RossumException(
+            f"Invalid JSON [{response.url}]: {response.text}", response=response
+        ) from e
 
 
 def get_text(response: Response) -> str:
     try:
         return response.text
     except ValueError as e:
-        raise RossumException(f"Invalid text [{response.url}]: {response.text}") from e
+        raise RossumException(
+            f"Invalid text [{response.url}]: {response.text}", response=response
+        ) from e
 
 
 def _encode_booleans(query: Optional[dict]) -> Optional[dict]:
